
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

extern "C"
{
#include <jpeglib.h>

#include "jpegdecoder_libjpeg.h"
#include "jpeg_header.h"
#include "jconfig.h"
}

#include <setjmp.h>




typedef struct
{
 struct          jpeg_source_mgr pub;
 unsigned char * inbuf;
 int             bufsize;
} my_source_mgr;
typedef my_source_mgr * my_src_ptr;

struct my_error_mgr {
  struct jpeg_error_mgr pub;    /* "public" fields */

  jmp_buf setjmp_buffer;        /* for return to caller */
};
typedef struct my_error_mgr * my_error_ptr;

METHODDEF(void)
my_error_exit (j_common_ptr cinfo)
{
  /* cinfo->err really points to a my_error_mgr struct, so coerce pointer */
  my_error_ptr myerr = (my_error_ptr) cinfo->err;

  /* Always display the message. */
  /* We could postpone this until after returning, if we chose. */
  //(*cinfo->err->output_message) (cinfo);

  /* Return control to the setjmp point */
  longjmp(myerr->setjmp_buffer, 1);
}

METHODDEF(void)
     my_displaymsg(j_common_ptr cinfo)
{
  /* cinfo->err really points to a my_error_mgr struct, so coerce pointer */
  my_error_ptr myerr = (my_error_ptr) cinfo->err;
  /* Return control to the setjmp point */
  longjmp(myerr->setjmp_buffer, 1);
}



METHODDEF(void) init_source (j_decompress_ptr cinfo)
{
}

METHODDEF(boolean) fill_input_buffer (j_decompress_ptr cinfo)
{
 my_src_ptr src = (my_src_ptr) cinfo->src;

 src->pub.next_input_byte = src->inbuf;
 src->pub.bytes_in_buffer = src->bufsize;
 return TRUE;
}

METHODDEF(void) skip_input_data (j_decompress_ptr cinfo, long num_bytes)
{

 my_src_ptr src = (my_src_ptr) cinfo->src;


 if (num_bytes > 0)
  {
   while (num_bytes > (long) src->pub.bytes_in_buffer)
    {
     num_bytes -= (long) src->pub.bytes_in_buffer;
     (void) fill_input_buffer(cinfo);
    }
   src->pub.next_input_byte += (size_t) num_bytes;
   src->pub.bytes_in_buffer -= (size_t) num_bytes;
  }
}

METHODDEF(void) term_source (j_decompress_ptr cinfo) { }


GLOBAL(void) jpeg_buf_src ( j_decompress_ptr cinfo, unsigned char * inbuf,int bufsize )

{
 my_src_ptr src;
 if (cinfo->src == NULL) cinfo->src= (jpeg_source_mgr*)malloc( sizeof( my_source_mgr ) );
 src = (my_src_ptr) cinfo->src;
 src->pub.init_source = init_source;
 src->pub.fill_input_buffer = fill_input_buffer;
 src->pub.skip_input_data = skip_input_data;
 src->pub.resync_to_restart = jpeg_resync_to_restart;
 src->pub.term_source = term_source;
 src->inbuf = (unsigned char*)inbuf;
 src->bufsize=bufsize;
 src->pub.bytes_in_buffer = 0;
 src->pub.next_input_byte = NULL;
}

extern "C"
{

void create_jpeg_from_data (u_int8_t * dst, u_int8_t * src, int qIndex,
		    int w, int h, u_int8_t format, int o_size,
		    int *size, int omit_huffman_table)
{

	int i = 0;
	u_int8_t *start;
	u_int8_t value, val1, val2, val3;

	start = dst;
	/* copy the header from the template */
	 memcpy (dst, GsmartJPGDefaultHeaderPart1,
		  GSMART_JPG_DEFAULT_HEADER_PART1_LENGTH);

	/* modify quantization table */
	memcpy (dst + 7, GsmartQTable[qIndex * 2], 64);
	memcpy (dst + 72, GsmartQTable[qIndex * 2 + 1], 64);

	dst += GSMART_JPG_DEFAULT_HEADER_PART1_LENGTH;

	/* copy Huffman table */
	if (!omit_huffman_table)
	{
		memcpy (dst, GsmartJPGDefaultHeaderPart2,
			GSMART_JPG_DEFAULT_HEADER_PART2_LENGTH);
		dst += GSMART_JPG_DEFAULT_HEADER_PART2_LENGTH;
	}
		memcpy (dst, GsmartJPGDefaultHeaderPart3,
			GSMART_JPG_DEFAULT_HEADER_PART3_LENGTH);

	/* modify the image width, height */
	*(dst + 8) = w & 0xFF;	//Image width low byte
	*(dst + 7) = w >> 8 & 0xFF;	//Image width high byte
	*(dst + 6) = h & 0xFF;	//Image height low byte
	*(dst + 5) = h >> 8 & 0xFF;	//Image height high byte

	/* set the format */
	*(dst + 11) = format;

	/* point to real JPG compress data start position and copy */
	dst += GSMART_JPG_DEFAULT_HEADER_PART3_LENGTH;
	/* copy data since we found a end of frame 0x00 0x00 0x00 */
	/* seem to work with all my test but ... mxhaard 5 06 2003 */
	for (i = 0; i < o_size; i++) {
		value = *(src + i) & 0xFF;
		val1 = *(src + i + 1) & 0xFF;
		val2 = *(src + i + 2) & 0xFF;
		val3 = *(src + i + 2) & 0xFF;
		*(dst) = value;
		dst++;
		if ((value + val1 + val2 + val3) == 0x00)
			break;
		if (value == 0xFF) {
			*(dst) = 0x00;
			dst++;
		}
	}

	/* Add end of image marker */
	*(dst++) = 0xFF;
	*(dst++) = 0xD9;

	*size = dst - start;
}



static struct jpeg_decompress_struct cinfo;
static unsigned char* data_with_header;
/* Output row buffer */
static JSAMPLE **buffer, *rowbuf;
static int row_stride;
static struct my_error_mgr jerr;

void close_libjpeg_decoder (void)
{
	free (data_with_header);
	free (rowbuf);
}

void init_libjpeg_decoder(int width, int height)
{
  data_with_header=(unsigned char*)malloc(width*height*bpp+300);
  rowbuf =  (unsigned char*)malloc(width*bpp+300);
  buffer = &rowbuf;
  row_stride=width*bpp;  /* physical row width in output buffer */
/*  int file_size; */

/*
  create_jpeg_from_data ((u_int8_t*)data_with_header,
			 (u_int8_t*)data_with_header, 1, IMAGE_WIDTH,
			 IMAGE_HEIGHT, 0x22, 100000, &file_size, 0);



  // Step 1: allocate and initialize JPEG decompression object
  // We set up the normal JPEG error routines, then override error_exit.
  cinfo.err = jpeg_std_error(&jerr.pub);
  jerr.pub.error_exit = my_error_exit;
  jerr.pub.output_message = my_displaymsg;
  // Establish the setjmp return context for my_error_exit to use.
  if (setjmp(jerr.setjmp_buffer)) {
    // If we get here, the JPEG code has signaled an error.
    // We need to clean up the JPEG object, close the input file, and return.
    jpeg_destroy_decompress(&cinfo);
    return;
  }
  // Now we can initialize the JPEG decompression object.
  jpeg_create_decompress(&cinfo);

  // Step 2: specify data source (eg, a file)
  jpeg_buf_src(&cinfo, data_with_header, file_size);

  // Step 3: read file parameters with jpeg_read_header()
  (void) jpeg_read_header(&cinfo, TRUE);
*/


  //jpeg_create_decompress(&cinfo);
  //jpeg_read_header(&cinfo,TRUE);
  //jpeg_start_decompress(&cinfo);
  //cinfo.output_width=IMAGE_WIDTH;
  //cinfo.output_height=IMAGE_HEIGHT;
  //jpeg_finish_decompress(&cinfo);

}

void libjpeg_decode_without_header (unsigned char *pdest, void *psrc,
					    int width, int height)
{
	int file_size;

	//memcpy(data_with_header,psrc,width*height*bpp);

	create_jpeg_from_data ((u_int8_t *) data_with_header,
			       (u_int8_t *) psrc, 2, width, height,
			       0x22, 100000, &file_size, 0);
	libjpeg_decode (pdest, data_with_header, file_size, width,
				height);
}

void libjpeg_decode(unsigned char* pdest, void* psrc, int file_size,
		  int width, int height)
{


  // Step 1: allocate and initialize JPEG decompression object
  // We set up the normal JPEG error routines, then override error_exit.
  cinfo.err = jpeg_std_error(&jerr.pub);
  jerr.pub.error_exit = my_error_exit;
  jerr.pub.output_message = my_displaymsg;
  // Establish the setjmp return context for my_error_exit to use.
  if (setjmp(jerr.setjmp_buffer)) {
    // If we get here, the JPEG code has signaled an error.
    // We need to clean up the JPEG object, close the input file, and return.
    jpeg_destroy_decompress(&cinfo);
    return;
  }
  // Now we can initialize the JPEG decompression object.
  jpeg_create_decompress(&cinfo);

  // Step 2: specify data source (eg, a file)
  jpeg_buf_src (&cinfo, (unsigned char *) psrc, file_size);
  // Step 3: read file parameters with jpeg_read_header()
  
  jpeg_read_header(&cinfo, TRUE);

  // Step 4: set parameters for decompression

  /* Step 5: Start decompressor */
  jpeg_start_decompress(&cinfo);

  /* JSAMPLEs per row in output buffer */
  row_stride = cinfo.output_width * cinfo.output_components;

  /* Make a one-row-high sample array that will go away when done with image */
  /*  buffer = (*cinfo.mem->alloc_sarray)
                ((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);
  */
  /* Step 6: while (scan lines remain to be read) */
  /*           jpeg_read_scanlines(...); */
  /* Here we use the library's state variable cinfo.output_scanline as the
   * loop counter, so that we don't have to keep track ourselves.
   */
  while (cinfo.output_scanline < cinfo.output_height) {
    int x;
    /* jpeg_read_scanlines expects an array of pointers to scanlines.
     * Here the array is only one element long, but you could ask for
     * more than one scanline at a time if that's more convenient.
     */
    
    if (bpp==3) {
    jpeg_read_scanlines(&cinfo,(JSAMPLE**)&pdest, 1);
    for(x=0;x<width*bpp;x+=bpp) {
      register char t;
      t=pdest[x+0];
      pdest[x+0]=pdest[x+2];
      pdest[x+2]=t;
      }
      pdest+=row_stride;
    } else {
    jpeg_read_scanlines(&cinfo,(JSAMPLE**)buffer, 1);
    unsigned char *pbuf=(unsigned char*)buffer[0];
    unsigned int v;
    for(x=0;x<width;x++) {
      v=(*((unsigned int*)pbuf));
      *((unsigned int*)pdest) =  
	((v&0xFF00)) | ((v&0xFF000000)>>8) | ((v&0xFF0000)>>16);
      pdest+=4;
      pbuf+=3;
    }
   
   } 
  }

  /* Step 7: Finish decompression */
  jpeg_finish_decompress(&cinfo);
  /* We can ignore the return value since suspension is not possible
   * with the stdio data source.
   */

  /* Step 8: Release JPEG decompression object */
  /* This is an important step since it will release a good deal of memory. */
  jpeg_destroy_decompress(&cinfo);
  /* After finish_decompress, we can close the input file.
   * Here we postpone it until after no more JPEG errors are possible,
   * so as to simplify the setjmp error logic above.  (Actually, I don't
   * think that jpeg_destroy can do an error exit, but why assume anything...)
   */

  /* At this point you may want to check to see whether any corrupt-data
   * warnings occurred (test whether jerr.pub.num_warnings is nonzero).
   */


}

} /* end extern C */
