/****************************************************************************
#	 	spcaview: Sdl video recorder and viewer with sound.         #
#This package work with the spca50x based webcam with the raw jpeg feature. #
#All the decoding is in user space with the help of libjpeg.                #
#.                                                                          #
# 		Copyright (C) 2003 2004 2005 Michel Xhaard                  #
#                                                                           #
# This program is free software; you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by      #
# the Free Software Foundation; either version 2 of the License, or         #
# (at your option) any later version.                                       #
#                                                                           #
# This program is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of            #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             #
# GNU General Public License for more details.                              #
#                                                                           #
# You should have received a copy of the GNU General Public License         #
# along with this program; if not, write to the Free Software               #
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA #
#                                                                           #
****************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <string.h>
#include <SDL/SDL.h>
#include <SDL/SDL_thread.h>
#include <SDL/SDL_audio.h>
#include <SDL/SDL_timer.h>
#include <linux/videodev.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <errno.h>
#include <fcntl.h>
#include <time.h>
#include <sys/time.h>
#include <signal.h>

#include "jpegdecoder_nolibjpeg.h"
#include "jpegdecoder_libjpeg.h"
#include "jconfig.h"

#include "avilib.h"
#include "dpsh.h"

#define WAVE_AUDIO_PCM 1
#define CHANNELS 1
#define AUDIO_SAMPLERATE		22050 // 44100 // 11025 // 8000 // 22050
#define AUDIO_SAMPLES			512 // 1024
#define AUDIO_FORMAT			AUDIO_S16
#define AUDIOSIZE 2
#define NUM_CHUNCK 64 // 256 a large ring buffer 
#define MAXBUFFERSIZE (AUDIO_SAMPLES*AUDIOSIZE*NUM_CHUNCK)
#define HELPBUFFERSIZE (AUDIO_SAMPLES*AUDIOSIZE)

// INIT_BRIGHT is the initial brightness level
// BRIGHTMEAN is the goal mean brightness for the 
// auto adjustment.
// SPRING_CONSTANT determines the restoring force
// for the auto adjust
// BRIGHT_WINDOW is the window around the goal mean
// in which no adjustment is to occur
#define V4L_BRIGHT_MIN 0
#define V4L_BRIGHT_MAX 50000 //65000
#define BRIGHTMEAN 128 //145
#define SPRING_CONSTANT 25
#define BRIGHTWINDOW 10

char version[] = "Spcaview version:0.41 date: 1:01:2005 (C) mxhaard@magic.fr\0";
typedef short indata;

SDL_Surface *pscreen;
SDL_Event sdlevent;
avi_t *out_fd;
u_int8_t *jpegData = NULL;	// Till lesson a good pointer is a pointer NULL
unsigned char *dpshDest = NULL;
int jpegSize;
unsigned char *p = NULL;
unsigned char *pp = NULL;
int image_width = IMAGE_WIDTH;
int image_height = IMAGE_HEIGHT;
int fullsize = 0;
int autobright = 0;
int interval = 0;
int pictFlag = 0;
/* sound buffer */
struct Rbuffer {
	Uint8 buffer[MAXBUFFERSIZE];
	Uint8 helpbuffer[HELPBUFFERSIZE];
	int ptread ;
	int ptwrite;
	};
	
static Uint32 callback_timer (Uint32 interval)
{
	pictFlag =1;
	return interval;
}
void callback_record(void *userdata, Uint8 *stream, int len)
{
	int part1 =0;
	int part2 =0;
 	struct Rbuffer *mybuf=(struct Rbuffer*) userdata;
 	if ((mybuf->ptwrite + len) > MAXBUFFERSIZE){
 		part1=MAXBUFFERSIZE-mybuf->ptwrite;
		part2 = len-part1;
		memcpy (mybuf->buffer+mybuf->ptwrite,stream,part1);
 		memcpy (mybuf->buffer,stream+part1,part2);
 	} else {
 		memcpy(mybuf->buffer+mybuf->ptwrite,stream,len);
 	}
 	mybuf->ptwrite = (mybuf->ptwrite +len) % MAXBUFFERSIZE;
 //fprintf(stderr,"callback_record(%p,%d,%d %d)\n",userdata,mybuf->ptwrite,part1,part2);
 
}
void callback_play(void *userdata, Uint8 *stream, int len)
{
	struct Rbuffer *mybuf=(struct Rbuffer*) userdata;
	
	int ptwrite=mybuf->ptwrite ;
	int ptread =mybuf->ptread;
	
	int part1 =0;
	int part2 =0;
	int delta =0;
	delta = ptwrite-ptread;
	
		if(delta < 0){
			if (((MAXBUFFERSIZE + delta) >= len)){
				
	 			part1=MAXBUFFERSIZE-ptread;
				if (part1 >=len){
				part1 = len;
				part2 = 0 ;
				} else {
	 			part2 = len-part1;
				}
				//fprintf(stderr,"callback_playT part1 %d part2 %d \n",part1,part2); 
			 	 memcpy (mybuf->helpbuffer,mybuf->buffer+ptread,part1);
				if (part2)
			 	    memcpy (mybuf->helpbuffer+part1,mybuf->buffer,part2);
			 	SDL_MixAudio(stream,mybuf->helpbuffer, len, SDL_MIX_MAXVOLUME);
				mybuf->ptread = (mybuf->ptread +len) % MAXBUFFERSIZE;
 				//fprintf(stderr,"callback_playT ptread %d ptwrite %d \n",ptread,ptwrite); 
			 
			 } else {
			 	//fprintf(stderr,"Waiting for dataT to Play Ptread %dPtwrite %d !! \n",ptread,ptwrite);
			}
		} else {
			if (delta >= len){
				memcpy (mybuf->helpbuffer,mybuf->buffer+ptread,len);
				SDL_MixAudio(stream,mybuf->helpbuffer, len, SDL_MIX_MAXVOLUME);
				mybuf->ptread = (mybuf->ptread +len) % MAXBUFFERSIZE;
 				//fprintf(stderr,"callback_playF ptread %d ptwrite %d \n",ptread,ptwrite);
				} else {
	 			//fprintf(stderr,"Waiting for dataF to Play Ptread %d Ptwrite %d !! \n",ptread,ptwrite);
			}
		}
	
	
	
 	
}
	
static int clip_to(int x, int low, int high);
static int adjust_bright( int width, int height, const char *addr, int is_rgb, struct video_picture *videopict, int fd);

static int get_pic_mean ( int width, int height, const unsigned char *buffer,
			  int is_rgb,int startx, int starty, int endx, 
			  int endy );
static int get_jpegsize (unsigned char *buf ,int insize);
static int setVideoPict (struct video_picture *videopict, int fd);

static int isSpcaChip (const char *BridgeName);
/* return Bridge otherwhise -1 */
static int getStreamId (const char * BridgeName);
/* return Stream_id otherwhise -1 */
static int probeSize (const char *BridgeName, int *width, int *height); 
/* return 1->ok otherwhise -1 */

static int
readFrame (long i, struct Rbuffer *RingBuffer, int maxsound, int *audiolen,
	   int *soundbytes, int isaudio, int updown, int compress);

static void
getPictureName (char *Picture, char *videoName, char *plan, int frame,
		char *ext);
/* get picture name by frame */		
static void getPicture( unsigned char *src,int width,int height, int frame);
/* get picture name by date */
static void getPictureT( unsigned char *src,int width,int height, int frame);
static void
YUV420toRGB (unsigned char *src, unsigned char *dst, int width, int height,
	     int flipUV, int ColSpace);

static void
refresh_screen (unsigned char *src, unsigned char *dst, int format, int width,
		int height, int size, int method);

int
main (int argc, char *argv[])
{
	/* Timing value used for statistic */
	Uint32 total_decode_time = 0;
	Uint32 time = 0;
	Uint32 prevtime = 0;
	Uint32 synctime = 0;
	Uint32 decodetime = 0;
	double average_decode_time = 0;
	Uint32 intime = 0;
	Uint32 pictime = 0;
	Uint32 delaytime = 0;
	Uint32 compresstime = 0;
	/**********************************/
	/*        video structures       */
	int fd;
	const char *videodevice = NULL;
	/* default mmap */
	int grabMethod = 1;
	int i, j, k, nframes = 2000, f, status;
	struct video_mmap vmmap;
	struct video_capability videocap;
	int mmapsize;
	struct video_mbuf videombuf;
	struct video_picture videopict;
	struct video_window videowin;
	struct video_channel videochan;
	int norme = 0;
	int channel = 0;
	int format = VIDEO_PALETTE_RAW_JPEG;
	/**********************************/
	/*        avi parametres          */
	char *compressor; 
	double framerate;
	long audiorate;
	int audiobits;
	int audioformat;
	unsigned char *pFramebuffer;
	unsigned char *tmp = NULL;	// don't forget the Till lesson
	int ff;
	int framecount = 0;
	/*********************************/
	/*       file input output       */
	char *inputfile = NULL;
	char *outputfile = NULL;
	static char Picture[80];
	char fourcc[4] = "MJPG";
	FILE *foutpict;
	char *sizestring = NULL;
	int use_libjpeg = 1;
	char *separateur;
	char *mode = NULL;
	/*********************************/
	/*          Starting Flags       */
	int run = 1;
	int quit = 1;
	int videoOn = 1;
	int decodeOn =1 ;
	int statOn = 0;
	int initAudio = 0; //flag start audio
	int videocomp = 0;
	int method = 1;
	int streamid ;
	int isVideoChannel = 1;
	int frame_size = 0;
	int len = 0; 
	/*********************************/
	/* data for SDL_audioin && SDL_audio */
	SDL_AudioSpec spec, result;
  	SDL_AudioSpec expect;
  	struct Rbuffer RingBuffer;
  	int retry = 100;
  	RingBuffer.ptread =0;
  	RingBuffer.ptwrite =0;
  	int err = 0; 	
	int bytes_per_read =0;
	int audioout = 0;
	int audiolength = 0;
	int maxsound = 0;
	/*********************************/

	/* init default bytes per pixels for VIDEO_PALETTE_RAW_JPEG 	*/
	/* FIXME bpp is used as byte per pixel for the ouput screen	*/
	/* we need also a bpp_in for the input stream as spcaview   	*/
	/* can convert stream That will be a good idea to have 2 struct */
	/* with all global data one for input the other for output      */ 
	bpp = 3;
	printf(" %s \n",version);
	/* check arguments */
	for (i = 1; i < argc; i++) {
		/* skip bad arguments */
		if (argv[i] == NULL || *argv[i] == 0 || *argv[i] != '-') {
			continue;
		}
		if (strcmp (argv[i], "-d") == 0) {
			if (i + 1 >= argc) {
				printf ("No parameter specified with -d, aborting.\n");
				exit (1);
			}
			videodevice = strdup (argv[i + 1]);
		}
		if (strcmp (argv[i], "-n") == 0) {
			if (i + 1 >= argc) {
				printf ("No parameter specified with -n, aborting.\n");
				exit (1);
			}
			norme = atoi (argv[i + 1]);
			if (norme < 0 || norme > 4)
			printf ("Norme should be between 0..4 Read the readme !.\n");
		}
		if (strcmp (argv[i], "-c") == 0) {
			if (i + 1 >= argc) {
				printf ("No parameter specified with -c, aborting.\n");
				exit (1);
			}
			channel = atoi (argv[i + 1]);
			if (channel < 0 || channel > 9)
			printf ("Channel should be between [0..3] || [6..9] Read the readme !.\n");
		}
		if (strcmp (argv[i], "-v") == 0) {
			videoOn = 0;

		}
		if (strcmp (argv[i], "-j") == 0) {
			decodeOn = 0;

		}
		if (strcmp (argv[i], "-t") == 0) {
			statOn = 1;

		}
		if (strcmp (argv[i], "-z") == 0) {
			videocomp = 1;

		}
		if (strcmp (argv[i], "-b") == 0) {
		  autobright = 1;
		}
		if (strcmp (argv[i], "-f") == 0) {
			if (i + 1 >= argc) {
				printf ("No parameter specified with -f, aborting.\n");
				exit (1);
			}
			mode = strdup (argv[i + 1]);

			if (strncmp (mode, "r32", 3) == 0) {
				format = VIDEO_PALETTE_RGB32;
				bpp = 4;
				snprintf (fourcc, 5, "RGB4");

			} else if (strncmp (mode, "r24", 3) == 0) {
				format = VIDEO_PALETTE_RGB24;
				bpp = 3;
				snprintf (fourcc, 5, "RGB3");
			} else if (strncmp (mode, "r16", 3) == 0) {
				format = VIDEO_PALETTE_RGB565;
				bpp = 2;
				snprintf (fourcc, 5, "RGB2");
			} else if (strncmp (mode, "yuv", 3) == 0) {
				format = VIDEO_PALETTE_YUV420P;
				bpp = 3;
				snprintf (fourcc, 5, "I420");
			} else if (strncmp (mode, "jpg", 3) == 0) {
				format = VIDEO_PALETTE_JPEG;
				bpp = 3;
				snprintf (fourcc, 5, "MJPG");
			}else {
				format = VIDEO_PALETTE_RAW_JPEG;
				bpp = 3;
				snprintf (fourcc, 5, "MJPG");
			}

		}

		if (strcmp (argv[i], "-i") == 0) {
			if (i + 1 >= argc) {
				printf ("No parameter specified with -i, aborting.\n");
				exit (1);
			}
			inputfile = strdup (argv[i + 1]);
		}

		if (strcmp (argv[i], "-g") == 0) {
			/* Ask for read instead default  mmap */
			grabMethod = 0;
		}

		if (strcmp (argv[i], "-a") == 0) {
			if (i + 1 >= argc) {
				printf ("No parameter specified with -a, aborting.\n");
				exit (1);
			}
			audioout = atoi (argv[i + 1]);
			if ((audioout < 0) || (audioout > 2)) {
				audioout = 0;
			}
			printf ("audio channel %d\n", audioout);
		}

		if (strcmp (argv[i], "-o") == 0) {
			if (i + 1 >= argc) {
				printf ("No parameter specified with -o, aborting.\n");
				exit (1);
			}
			outputfile = strdup (argv[i + 1]);
		}
		/* custom ? */
		if (strcmp (argv[i], "-s") == 0) {
			if (i + 1 >= argc) {
				printf ("No parameter specified with -s, aborting.\n");
				exit (1);
			}

			sizestring = strdup (argv[i + 1]);

			image_width = strtoul (sizestring, &separateur, 10);
			if (*separateur != 'x') {
				printf ("Error in size use -s widthxheight \n");
				exit (1);
			} else {
				++separateur;
				image_height =
					strtoul (separateur, &separateur, 10);
				if (*separateur != 0)
					printf ("hmm.. dont like that!! trying this height \n");
				printf (" size width: %d height: %d \n",
					image_width, image_height);
			}
		}
		if (strcmp (argv[i], "-p") == 0) { 		  
			if (argv[i + 1]) interval = 1000* (atoi(argv[i + 1])); // timer works on ms
 		  	else {
 		   	 	printf ("No parameter specified with -p, aborting.\n");
 		    		exit (1);
 		  	}

 		}

		/* large ? */
		if (strcmp (argv[i], "-l") == 0) {
			image_width = 640;
			image_height = 480;
		}
		


		if (strcmp (argv[i], "-h") == 0) {
			printf ("usage: spcaview [-h -d -n -c -v -j -z -g -l -s -a -f -b -t -p] [-i inputfile | -o outputfile]\n");
			printf ("-h	print this message \n");
			printf ("-d	/dev/videoX       use videoX device\n");
			printf ("-n	norme 0->pal 2->secam 1->ntsc \n");
			printf ("-c	input channels 0..9 exclude 4 and 5 \n");
			printf ("-v	disable video output for raw recording \n");
			printf ("-j	disable video output and userspace decoding \n");
			printf ("-z	compress yuv420p video output with DPSH encoder\n");
			printf ("-g	use read method for grab instead mmap \n");
			printf ("-l	default 320x480   use 640x480 \n");
			printf ("-s	widthxheight      use specified size \n");
			printf ("-a	default  0 -> nosound    1-> microphone 2-> mixer output\n");
			printf ("-f	video format  default rawjpg  others options are r16 r24 r32 yuv jpg \n");
			printf ("-b     enable automatic brightness adjustment \n");
			printf ("-t     print statistics \n");
			printf ("-p  x  getPicture every x seconds \n");
			exit (0);
		}
	}



	printf ("Initializing SDL.\n");
	/* Initialize defaults, Video and Audio */
	if ((SDL_Init (SDL_INIT_VIDEO | SDL_INIT_TIMER) == -1)) {
		printf ("Could not initialize SDL: %s.\n", SDL_GetError ());
		exit (-1);
	}

	/* Clean up on exit */
	atexit (SDL_Quit);

	printf ("SDL initialized.\n");

	if (inputfile) {
		/* that a is spcaview player */
		printf ("Using %s avifile \n", inputfile);
		if ((out_fd = AVI_open_input_file (inputfile, 1)) == NULL) {
			printf ("cannot open input file ? \n");
			exit (1);
		}
		/* get the parameters from the file */

		framecount = (int) AVI_video_frames (out_fd);

		image_width = (int) AVI_video_width (out_fd);
		image_height = (int) AVI_video_height (out_fd);
		printf ("frame: %d  width: %d height: %d \n",
			framecount, image_width, image_height);
		framerate = AVI_frame_rate (out_fd);
		printf (" framerate : %f \n", framerate);
		audiobits = AVI_audio_bits (out_fd);
		audiorate = AVI_audio_rate (out_fd);
		audioformat = AVI_audio_format (out_fd);
		printf ("audio rate: %d audio bits: %d audio format: %d\n",
			audiorate, audiobits, audioformat);

		compressor = AVI_video_compressor (out_fd);
		printf (" codec: %s \n", compressor);

		if (strncmp (compressor, "RGB4", 4) == 0) {
			format = VIDEO_PALETTE_RGB32;
			bpp = 4;
		} else if (strncmp (compressor, "RGB3", 4) == 0) {
			format = VIDEO_PALETTE_RGB24;
			bpp = 3;
		} else if (strncmp (compressor, "RGB2", 4) == 0) {
			format = VIDEO_PALETTE_RGB565;
			bpp = 2;
		} else if (strncmp (compressor, "I420", 4) == 0) {
			format = VIDEO_PALETTE_YUV420P;
			bpp = 3;
		} else if (strncmp (compressor, "MJPG", 4) == 0) {
			format = VIDEO_PALETTE_RAW_JPEG;
			bpp = 3;
		} else if (strncmp (compressor, "DPSH", 4) == 0) {
			format = VIDEO_PALETTE_YUV420P;
			videocomp = 1;
			frame_size = (image_width*image_height*3) >> 1;
			bpp = 3;
		} else {
			printf ("Cannot play this file try Mplayer or Xine \n");
			AVI_close (out_fd);
			exit (1);
		}
		/* Init the player */
		printf ("bpp %d format %d\n", bpp, format);
		if (format == VIDEO_PALETTE_RAW_JPEG)
			init_libjpeg_decoder (image_width, image_height);
		pscreen =
			SDL_SetVideoMode (image_width, image_height, bpp * 8,
					  SDL_DOUBLEBUF | SDL_SWSURFACE);
		if (pscreen == NULL) {
			printf ("Couldn't set %d*%dx%d video mode: %s\n",
				image_width, image_height, bpp * 8,
				SDL_GetError ());
			exit (1);
		}

		/* No Cursor in Player */
		SDL_ShowCursor (0);
		SDL_WM_SetCaption ("Spcaview Player ", NULL);

		p = pscreen->pixels;
		synctime = 1000 / framerate;
		printf ("Frame time %d ms\n", synctime);
		if (AVI_audio_channels (out_fd) == 1) {
			
			expect.format=AUDIO_FORMAT;
  			expect.freq=audiorate;
  			expect.callback=callback_play;
  			expect.samples=AUDIO_SAMPLES;
  			expect.channels=CHANNELS;
  			expect.userdata=(void*)&RingBuffer;
			if(SDL_OpenAudio(&expect,NULL)<0)
  				{
    				fprintf(stderr,"Couldn't open audio output device: %s\n",SDL_GetError());
    				SDL_CloseAudio();
    				SDL_Quit();
  				}
			/* Init the sound player */
			bytes_per_read = 2 * audiorate / framerate;
			maxsound = audiolength = AVI_audio_bytes (out_fd);
			//printf ("bytes to read %d \n",bytes_per_read);
			initAudio =1;
			/* mute if not a raw PCM file */
			if (audioformat == 1)
				audioout = 1;
		}
		i = 0;
		dpshDest = malloc(frame_size);
		do {
			if (run) {
				intime = SDL_GetTicks ();
				/* read frame read video and audio */
				err =readFrame (i,&RingBuffer, maxsound,
					   &audiolength, &bytes_per_read,
					   audioout, 1,videocomp);

				//printf("readframe %d audiolength %d position errors %d \n",i,audiolength,err);
				
				if (initAudio ){
					initAudio = 0;
					SDL_PauseAudio(0); // start play
				}
				refresh_screen ( jpegData, p, format, image_width,image_height,jpegSize,0);

				SDL_Flip (pscreen);	//switch the buffer and update screen
				pictime = SDL_GetTicks () - intime;
				//printf(" get pictimes %d expected %d  \n",pictime,synctime);
				if (pictime < synctime)
					SDL_Delay (synctime - pictime);
				i++;
			}
			fflush (stdout);
			if (SDL_PollEvent (&sdlevent) == 1) {
				switch (sdlevent.type) {
					case SDL_KEYDOWN:
						switch (sdlevent.key.keysym.sym) {
							case SDLK_UP:
								initAudio = 1;
								run = 1;
								break;
							case SDLK_DOWN:
								SDL_PauseAudio(1); // stop play
								run = 0;
								break;
							case SDLK_RIGHT:
								if (i <
								    (framecount
								     - 1)) {
									i++;

									err = readFrame
										(i,
										 &RingBuffer,
										 maxsound,
										 &audiolength,
										 &bytes_per_read,
										 audioout,
										 1,videocomp);
									printf ("scanning up frame %d\n", i);
									refresh_screen
										(jpegData,
										 p,
										 format,
										 image_width,
										 image_height,
										 jpegSize,0);

									SDL_Flip (pscreen);	//switch the buffer and update screen
								}
								break;
							case SDLK_LEFT:
								if (i > 1) {
									i--;

									err = readFrame
										(i,
										 &RingBuffer,
										 maxsound,
										 &audiolength,
										 &bytes_per_read,
										 audioout,
										 0,videocomp);
									printf ("scanning down frame %d\n", i);
									refresh_screen
										(jpegData,
										 p,
										 format,
										 image_width,
										 image_height,
										 jpegSize,0);

									SDL_Flip (pscreen);	//switch the buffer and update screen
								}
								break;
							case SDLK_s:
								if (run == 0) {
									switch (format) {
										case VIDEO_PALETTE_RAW_JPEG:{
												getPictureName
													(Picture,
													 inputfile,
													 "J-",
													 i,
													 "jpg");
												printf ("\n Got: picture name: %s\n", Picture);
												if (Picture) {
													int fdpict = open (Picture, O_CREAT | O_RDWR, S_IRWXU | S_IRWXO);

													write (fdpict, jpegData, jpegSize);
													close (fdpict);
												}
											}
											break;
										case VIDEO_PALETTE_RGB32:{
												/* FIXME */
												/* no found a good picture format with32 bits */
												printf ("\n No picture available for that mode: %s\n", Picture);
											}
											break;
										case VIDEO_PALETTE_RGB24:{
												getPictureName
													(Picture,
													 inputfile,
													 "R-",
													 i,
													 "pnm");
												printf ("\n Got: picture name: %s\n", Picture);
												if (Picture) {
													foutpict = fopen (Picture, "wb");
													fprintf (foutpict, "P6\n%d %d\n255\n", image_width, image_height);
													for (k = 0; k < image_width * image_height * bpp; k += 3) {
														j = jpegData[k];
														jpegData[k] = jpegData[k + 2];
														jpegData[k + 2] = j;
													}
													fwrite (jpegData, sizeof (char), image_width * image_height * bpp, foutpict);
													fclose (foutpict);
												}
											}
											break;

										case VIDEO_PALETTE_YUV420P:{
												tmp = malloc (image_width * image_height * bpp);

												YUV420toRGB
													((unsigned char *) jpegData, (unsigned char *) tmp, image_width, image_height, 1, 0);
												getPictureName
													(Picture,
													 inputfile,
													 "Y-",
													 i,
													 "pnm");
												printf ("\n Got: picture name: %s\n", Picture);
												if (Picture) {
													foutpict = fopen (Picture, "wb");
													fprintf (foutpict, "P6\n%d %d\n255\n", image_width, image_height);
													fwrite (tmp, sizeof (char), image_width * image_height * bpp, foutpict);

													fclose (foutpict);
												}
												free (tmp);
											}
											break;
										default:
											break;
									}
								}
								break;
							case SDLK_SPACE:
								if (fullsize ==
								    0) {
									pscreen = SDL_SetVideoMode (image_width, image_height, bpp * 8, SDL_DOUBLEBUF | SDL_FULLSCREEN | SDL_SWSURFACE);
									fullsize = 1;
								} else {
									pscreen = SDL_SetVideoMode (image_width, image_height, bpp * 8, SDL_DOUBLEBUF | SDL_SWSURFACE);
									fullsize = 0;
								}
								if (pscreen ==
								    NULL) {
									printf ("Couldn't set %d*%dx%d video mode: %s\n", image_width, image_height, bpp * 8, SDL_GetError ());
									exit (1);
								}
								/* No Cursor in Player */
								SDL_ShowCursor
									(0);
								break;
							default:
								printf ("\nStop asked\n");
								quit = 0;
								break;
						}
						break;
					case SDL_QUIT:
						quit = 0;
						break;
				}
			}
		}
		while (i < framecount && quit == 1);
		free (jpegData);
		free(dpshDest);
		close_libjpeg_decoder ();
		if (audioout) {
		
		while ((RingBuffer.ptread != RingBuffer.ptwrite) && retry--){
 			SDL_Delay(10);
  			//fprintf(stderr,"Waiting .. stop the player %d\n",retry);
 			}
			 SDL_CloseAudio(); // stop play
			
			audioout = 0;
		
		}
		AVI_close (out_fd);

	} else {
		/* spcaview grabber */		
		/* validate parameters */
		printf ("bpp %d format %d\n", bpp, format);
		if(!videoOn) {
		/* acquire raw data force palette to raw */
		printf ("VideoOn = 0\n");
			snprintf (fourcc, 5, "RAWD");
			format = VIDEO_PALETTE_RAW_JPEG;
			}
		if(videocomp) {
		/* acquire yuv420p and compress diff pixels static huffman */
			snprintf (fourcc, 5, "DPSH");
			format = VIDEO_PALETTE_YUV420P;
			}
		if (videodevice == NULL || *videodevice == 0) {
			videodevice = "/dev/video0";
		}
		printf ("Using video device %s.\n", videodevice);
		printf ("Initializing v4l.\n");
		// Init the two decoders
		init_libjpeg_decoder (image_width, image_height);
		init_nolibjpeg_decoder ();
		//v4l init
		if ((fd = open (videodevice, O_RDWR)) == -1) {
			perror ("ERROR opening V4L interface \n");
			exit (1);
		}
		printf("**************** PROBING CAMERA *********************\n");
		if (ioctl (fd, VIDIOCGCAP, &videocap) == -1) {
			printf ("wrong device\n");
			exit (1);
		}
		
		printf("Camera found: %s \n",videocap.name);
		
		if (ioctl (fd, VIDIOCGCHAN, &videochan) == -1) {
			printf ("Hmm did not support Video_channel\n");
			isVideoChannel = 0;
		}
		if (isVideoChannel){
			videochan.norm = norme;
			videochan.channel = channel;
			if (ioctl (fd, VIDIOCSCHAN, &videochan) == -1) {
				printf ("ERROR setting channel and norme \n");
				exit (1);
			}
			/************ just to be sure *************/
			if (ioctl (fd, VIDIOCGCHAN, &videochan) == -1) {
				printf ("wrong device\n");
				exit (1);
			}
			printf("Bridge found: %s \n",videochan.name);
			streamid = getStreamId (videochan.name);
		
			if (streamid >= 0){
				printf("StreamId: %s Camera\n",Plist[streamid].name);
			} else {
				printf("StreamId: %d Unknow Camera\n",streamid);
			} 
			/* test jpeg capability if not and jpeg ask without raw data 
			set default format to YUVP */
			if ((format == VIDEO_PALETTE_RAW_JPEG || format == VIDEO_PALETTE_JPEG )&& streamid != JPEG && videoOn) {
				printf ("Camera unable to stream in JPEG mode switch to YUV420P\n");
				format = VIDEO_PALETTE_YUV420P;
			}
			if(probeSize(videochan.name,&image_width,&image_height) < 0)
				printf("unable to probe size !!\n");
			}
		printf("*****************************************************\n");
		/* Init grab method mmap */
		if (grabMethod) {
			printf(" grabbing method default MMAP asked \n");
			// MMAP VIDEO acquisition
			memset (&videombuf, 0, sizeof (videombuf));
			if (ioctl (fd, VIDIOCGMBUF, &videombuf) < 0) {
				perror (" init VIDIOCGMBUF FAILED\n");
			}
			printf ("VIDIOCGMBUF size %d  frames %d  offets[0]=%d offsets[1]=%d\n", videombuf.size, videombuf.frames, videombuf.offsets[0], videombuf.offsets[1]);
			pFramebuffer =
				(unsigned char *) mmap (0, videombuf.size,
							PROT_READ | PROT_WRITE,
							MAP_SHARED, fd, 0);
			mmapsize = videombuf.size;
			vmmap.height = image_height;
			vmmap.width = image_width;
			vmmap.format = format;
			for (f = 0; f < videombuf.frames; f++) {
				vmmap.frame = f;
				if (ioctl (fd, VIDIOCMCAPTURE, &vmmap)) {
					perror ("cmcapture");
				}
			}

			vmmap.frame = 0;
			/* Compute the estimate frame size we expect that jpeg compress factor 10 
			if ((format == VIDEO_PALETTE_RAW_JPEG) && videoOn)
				frame_size =
					videombuf.size / (10 *
							  videombuf.frames);
			else
				frame_size = videombuf.size / videombuf.frames;
			*/
		} else {
			/* read method */
			printf(" grabbing method READ asked \n");
			if (ioctl (fd, VIDIOCGWIN, &videowin) < 0)
				perror ("VIDIOCGWIN failed \n");
			videowin.height = image_height;
			videowin.width = image_width;
			if (ioctl (fd, VIDIOCSWIN, &videowin) < 0)
				perror ("VIDIOCSWIN failed \n");
			printf ("VIDIOCGWIN height %d  width %d \n",
				videowin.height, videowin.width);
			
		}
		switch (format) {
				case VIDEO_PALETTE_JPEG:{
					frame_size = image_width * image_height;
				}
				break;
				case VIDEO_PALETTE_RAW_JPEG:{
					if ( videoOn)
						frame_size =
							image_width *
							image_height;
					else
					frame_size = image_width *
							image_height * 3;
					}
					break;
				case VIDEO_PALETTE_YUV420P:{
						frame_size =
							(image_width *
							 image_height * 3) >> 1;
					}
					break;
				case VIDEO_PALETTE_RGB565:
				case VIDEO_PALETTE_RGB24:
				case VIDEO_PALETTE_RGB32:{
						frame_size =
							image_width *
							image_height * bpp;
					}
					break;
				default:
					break;

			}
		/* struct video_picture VIDIOCGPICT VIDIOCSPICT */
		if (ioctl (fd, VIDIOCGPICT, &videopict) < 0) {
			perror ("Couldnt get videopict params with VIDIOCGPICT\n");
		}
		printf ("VIDIOCGPICT\n");
		printf("brightnes=%d hue=%d color=%d contrast=%d whiteness=%d \n" ,
		 videopict.brightness, videopict.hue, videopict.colour, videopict.contrast, videopict.whiteness);
		printf("depth=%d palette=%d\n", videopict.depth, videopict.palette);
		
		videopict.palette = format;
		videopict.depth = bpp * 8;
		//videopict.brightness = INIT_BRIGHT;
		sleep (1);
		setVideoPict (&videopict, fd);


		/*
		 * Initialize the display 
		 */
if ( decodeOn && videoOn )
			{	/* Display data */
		pscreen =
			SDL_SetVideoMode (image_width, image_height, bpp * 8,
					  SDL_SWSURFACE);
		if (pscreen == NULL) {
			printf ("Couldn't set %d*%dx%d video mode: %s\n",
				image_width, image_height, bpp * 8,
				SDL_GetError ());
			exit (1);
		}
//		SDL_WM_SetCaption ("Spcaview Grabber ", NULL);
SDL_WM_SetCaption (videocap.name, NULL);
}
		printf ("\n");
		time = SDL_GetTicks ();
		prevtime = time;

		/* If need open the avi file and alloc jpegbuffer */
		if (outputfile) {
			jpegData = malloc (frame_size);
			dpshDest = malloc(frame_size);
			if ((out_fd =
			     AVI_open_output_file (outputfile)) == NULL) {
				printf ("cannot open write file ? \n");
				exit (1);
			}
			AVI_set_video (out_fd, image_width, image_height, 20,
				       fourcc);
			if (audioout) {
			/* init the sound recorder */
			spec.format=AUDIO_FORMAT;
  			spec.freq=AUDIO_SAMPLERATE;
  			spec.callback=callback_record;
  			spec.samples=AUDIO_SAMPLES;
  			spec.channels=1;
  			spec.userdata=(void*)&RingBuffer;
			if(SDL_InitAudioIn()<0)
  				{
   				 fprintf(stderr,"Couldn't initialize SDL_AudioIn: %s\n",SDL_GetError());
    				 SDL_Quit();
   				 return(2);
  				}
			 if(SDL_OpenAudioIn(&spec,&result)<0)
  				{
   				 fprintf(stderr,"Couldn't open audio input device: %s\n",SDL_GetError());
   				 SDL_CloseAudioIn();
    				 SDL_Quit();
  				}
			AVI_set_audio (out_fd, spec.channels, spec.freq,
					       16, WAVE_AUDIO_PCM);
			//printf ("audio record setting channel %d frequency %d format %d",spec.channels, spec.freq, WAVE_AUDIO_PCM);
			
			// SDL_PauseAudioIn(0); // start record		
			initAudio = 1;
			}
		}

		/* Allocate tmp buffer for one frame. */

		tmp = (unsigned char*)malloc (frame_size);
	 	if (interval && decodeOn && videoOn){
			// set_timer(interval);
			SDL_SetTimer((Uint32) interval,callback_timer);
			 }
		i = 0;
		
		while (run) {
			memset(tmp,0x00,frame_size);
			intime = SDL_GetTicks ();
			pictime = intime - delaytime;
			delaytime = intime;
			/* Try to synchronize sound with frame rate */
			if (initAudio && i > 9){
					initAudio = 0;
					SDL_PauseAudioIn(0); // start record
				}
			/* compute bytes sound */
			if (pictime < 100) {
				
				bytes_per_read =
					((AUDIO_SAMPLERATE / 1000) * 2 * pictime);
			}
			i++;
			if (grabMethod) {
				ff = vmmap.frame;
				if (ioctl (fd, VIDIOCSYNC, &ff) < 0) {
					perror ("cvsync err\n");
				}
				vmmap.frame = ff;
				memcpy (tmp,
					pFramebuffer +
					videombuf.offsets[vmmap.frame],
					frame_size);
				if ((status =
				     ioctl (fd, VIDIOCMCAPTURE, &vmmap)) < 0) {
					perror ("cmcapture");
					printf (">>cmcapture err %d\n", status);
				}
				vmmap.frame =
					(vmmap.frame + 1) % videombuf.frames;
				
			} else {
				/* read method */
				len = read (fd, tmp, frame_size);
				// printf ("len %d asked %d \n",len,frame_size);
			}
			synctime = SDL_GetTicks ();
			if ((outputfile) && i > 10) {
				/* Output video frame with the good format */
				switch (format) {
					case VIDEO_PALETTE_JPEG:{
						int fs = get_jpegsize (tmp , frame_size);
							if (AVI_write_frame (out_fd,
							     		    (unsigned char *)
							     		     tmp,
							    		     fs) < 0)
									printf ("write error on avi out \n");
									method = 0;
						}
						break;
					case VIDEO_PALETTE_RAW_JPEG:{
							if (videoOn) {
								create_jpeg_from_data
									(jpegData, tmp,
								 	2, image_width,
									 image_height,
									 0x22,
									 frame_size,
									 &jpegSize, 0);
								if (AVI_write_frame
							    			(out_fd,
							     			(unsigned char *)
							     			jpegData,
							     			jpegSize) < 0)
									printf ("write error on avi out \n");
							} else {
								if (AVI_write_frame
							    			(out_fd,
							     			(unsigned char *)
							     			tmp,
							    			 frame_size) < 0)
									printf ("write error on avi out \n");
							}
						}
						break;
					case VIDEO_PALETTE_YUV420P:{
							if ( videocomp ) {

								memcpy( jpegData,tmp,frame_size);
								jpegSize=frame_size;
								/*jpegData is destroy here */
								dpsh_yuv_encode(jpegData, dpshDest, &jpegSize);

							// printf ("write DPSH Size %d \n", jpegSize);
							if (AVI_write_frame
									 (out_fd,
									(unsigned char *)
									dpshDest, jpegSize) < 0)
									printf ("write error on avi out \n");
							}else{
								if (AVI_write_frame
									 (out_fd,
									(unsigned char *)
									tmp,
									image_width *
									image_height *
									1.5) < 0)
									printf ("write error on avi out \n");
							}
						}
						break;
					case VIDEO_PALETTE_RGB565:
					case VIDEO_PALETTE_RGB24:
					case VIDEO_PALETTE_RGB32:{
							if (AVI_write_frame
							    (out_fd,
							     (unsigned char *)
							     tmp,
							     image_width *
							     image_height *
							     bpp) < 0)
								printf ("write error on avi out \n");
						}
						break;
					default:
						break;
				}
				/* write sound in avi */
				if (audioout) {
					//printf("bytes per read sound %d \n",bytes_per_read);
					SDL_LockAudio();
					int ptread = RingBuffer.ptread;
					int ptwrite = RingBuffer.ptwrite;
					SDL_UnlockAudio();
					if (ptwrite > ptread) {
					if (AVI_write_audio
						    (out_fd, (char *) RingBuffer.buffer+ptread,
						     ptwrite-ptread) < 0)
							printf (" write AVI error \n");
						 RingBuffer.ptread =  (RingBuffer.ptread + (ptwrite-ptread))%MAXBUFFERSIZE;	
					} else if (ptwrite < ptread ){
					if (AVI_write_audio
						    (out_fd, (char *) RingBuffer.buffer+ptread,
						     MAXBUFFERSIZE-ptread) < 0)
							printf (" write AVI error \n");
					if (AVI_write_audio
						    (out_fd, (char *) RingBuffer.buffer,
						     ptwrite) < 0)
							printf (" write AVI error \n");
					 RingBuffer.ptread =  (RingBuffer.ptread + ptwrite+(MAXBUFFERSIZE-ptread))%MAXBUFFERSIZE;		
					}
				
				}
				
			}

			compresstime =SDL_GetTicks ();

			
			
			if ( decodeOn && videoOn && (i > 10))
			{	/* Display data */
				p = pp = pscreen->pixels;
				if (autobright)
			  adjust_bright(image_width, image_height, pscreen->pixels, 1, &videopict, fd);
			  refresh_screen ( tmp, p, format, image_width,image_height,image_width*image_height*bpp,method);
			  decodetime = SDL_GetTicks ();

			SDL_UpdateRect (pscreen, 0, 0, 0, 0);	//update the entire screen	
			if (pictFlag){
				pictFlag = 0;
				getPictureT(p,image_width,image_height,i);
			}
			} else {
			
			decodetime = SDL_GetTicks ();
			}
			
			/* increment the real frame count and update statistics */
			framecount++;

			total_decode_time += decodetime - synctime;
			average_decode_time = total_decode_time / framecount;

			if (!autobright && statOn){
// \r
				printf ("frames:%04d pict:%03dms synch:%03dms write/comp:%03dms decode:%03dms display:%03dms \n",
				 framecount, pictime, synctime-intime, compresstime-synctime, decodetime-compresstime,SDL_GetTicks ()-decodetime); 
			fflush (stdout);	 
			}
			
			if (SDL_PollEvent (&sdlevent) == 1) {
				switch (sdlevent.type) {
					case SDL_KEYDOWN:
						switch (sdlevent.key.keysym.sym) {
							case SDLK_b:
								videopict.
									brightness
									-=
									0x100;
								setVideoPict
									(&videopict,
									 fd);
								break;
							case SDLK_n:
								videopict.
									brightness
									+=
									0x100;
								setVideoPict
									(&videopict,
									 fd);
								break;
							case SDLK_c:
								videopict.colour -= 0x200;	// = saturation ?
								setVideoPict
									(&videopict,
									 fd);
								break;
							case SDLK_s:
								getPicture(p,image_width,image_height,i);
								break;
							case SDLK_v:
								videopict.colour += 0x200;	// = saturation ?
								setVideoPict
									(&videopict,
									 fd);
								break;
							case SDLK_w:
								videopict.
									contrast -= 0x200;	// -contrast
								setVideoPict
									(&videopict,
									 fd);
								break;
							case SDLK_x:
								videopict.
									contrast += 0x200;	// +contrast
								setVideoPict
									(&videopict,
									 fd);
								break;
							default:
								printf ("\nStop asked\n");
								run = 0;
								break;
						}
						break;
					case SDL_QUIT:
						run = 0;
						break;
				}
			}

		}
		
			if (audioout) {
			
				SDL_CloseAudioIn(); //stop record 
				printf ("free sound buffer\n");
			
			}
			if (interval && decodeOn && videoOn){
			// set_timer(interval);
			SDL_SetTimer((Uint32) 0, NULL);
			 }
		time = SDL_GetTicks () - time;
		printf ("\nUsed %dms for %d images => %dms/image %dfps.\n",
			time, framecount, time / framecount,
			framecount * 1000 / time);


		printf ("Quiting SDL.\n");

		printf ("Decoded frames:%d Average decode time: %f\n",
			framecount, average_decode_time);
		if (grabMethod) {
			printf ("unmapping\n");
			munmap (pFramebuffer, mmapsize);
		}
		printf ("closing\n");
		close (fd);
		printf ("closed\n");
		if (outputfile) {
			out_fd->fps = (double) framecount *1000 / time;

			AVI_close (out_fd);
			printf ("close avi\n");
			if (audioout) {
			
				 SDL_QuitAudioIn();
				printf ("free sound buffer\n");
			
			}
			free (jpegData);
			free(dpshDest);
		}
		/* Shutdown all subsystems */
		free (tmp);
		close_libjpeg_decoder ();
		printf ("Quiting....\n");
	}
	SDL_Quit ();
	exit (0);
}

static int
 setVideoPict (struct video_picture *videopict, int fd)
{
	if (ioctl (fd, VIDIOCSPICT, videopict) < 0) {
		perror ("Couldnt get videopict params with VIDIOCSPICT\n");
		return -1;
	}
	printf ("VIDIOCSPICT\n");
	printf ("brightness=%d hue=%d color=%d contrast=%d whiteness=%d \n",
		videopict->brightness, videopict->hue,
		videopict->colour, videopict->contrast, videopict->whiteness);
		printf("depth=%d palette=%d \n",videopict->depth, videopict->palette);
	  
	return 0;
}

static int
 get_jpegsize (unsigned char *buf, int insize)
{
 int i; 	
 for ( i= 1024 ; i< insize; i++) {
 	if ((buf[i] == 0xFF) && (buf[i+1] == 0xD9)) return i+10;
 }
 return -1;
}

static int
 readFrame (long i,struct Rbuffer *RingBuffer, int maxsound, int *audiolen,
	   int *soundbytes, int isaudio, int updown, int compress )
{
	static int avipos = 0;
	int ptwrite;
	int ptread;
	int audiosize=0;
	int delta =0 ;
	/* got frame i and read the data */

	if (AVI_set_video_position (out_fd, i) < 0)
		printf ("error get video position\n");
		jpegSize = (int) AVI_frame_size (out_fd, i);
		//audiosize = (int)AVI_chunk_size (out_fd, i);
		audiosize = *soundbytes;
		avipos += audiosize;
		delta = ((i+1) * (*soundbytes))-avipos;
		//printf("audiosize delta %d \n",delta);
		jpegData = (unsigned char *) realloc (jpegData, (size_t) jpegSize);
	if (AVI_read_frame (out_fd, jpegData) < 0)
		printf ("error read frame\n");
	if (compress){
		/* decompress the video stream */
		dpsh_yuv_decode(jpegData , dpshDest, &jpegSize);
		/* restore jpegData size to the uncompressed mode */
		jpegData = (unsigned char *) realloc (jpegData, (size_t) jpegSize);
		memcpy(jpegData, dpshDest, jpegSize);
	}

	if (isaudio && (*audiolen > 0)) {
			SDL_LockAudio();
				ptread = RingBuffer->ptread;
				ptwrite = RingBuffer->ptwrite;
			SDL_UnlockAudio();
		if (updown) {
			/* upstream */
			/* read the audio Byte until no Byte */
			
					if (ptwrite + audiosize < MAXBUFFERSIZE) {

					if (AVI_read_audio
						    (out_fd, (char *) RingBuffer->buffer+ptwrite,
						     audiosize) < 0)
							 printf (" Read AVIchunck error \n");	
					} else {
					//printf ("part1 %d, part2 %d \n",MAXBUFFERSIZE-ptwrite,audiosize -( MAXBUFFERSIZE-ptwrite));

					if (AVI_read_audio
						    (out_fd, (char *) RingBuffer->buffer+ptwrite,
						     MAXBUFFERSIZE-ptwrite) < 0)
							 printf (" Read AVIpart1 error \n");
					if (AVI_read_audio
						    (out_fd, (char *) RingBuffer->buffer,
						     audiosize -( MAXBUFFERSIZE-ptwrite)) < 0)
							 printf (" Read AVIpart2 error \n");
		
					}
			SDL_LockAudio();	
				RingBuffer->ptwrite =  (RingBuffer->ptwrite + audiosize)%MAXBUFFERSIZE;
			SDL_UnlockAudio();
			
			*audiolen -= audiosize;
			
			

		} else {
			/* down stream */
			/* rewind one sound fragment */
			*audiolen += *soundbytes ;
			avipos = (maxsound - *audiolen);
			if (avipos >= 0)
				if (AVI_set_audio_position (out_fd, (long)avipos) < 0)
					printf ("error read soundframe down \n");
			//printf("audiolenth : %d soundbytes %d aviposition %d\n",*audiolen,audiosize,avipos);


		}

	}
return delta;
}

static void
 getPictureName (char *Picture, char *videoName, char *plan, int frame,
		char *ext)
{
	char tamp[80];
	char myPicture[80];
	int namesize;
	memset (myPicture,'\0',sizeof(myPicture));
	memset (tamp,'\0',sizeof(tamp));
	namesize = strlen (videoName) + strlen (plan);
	snprintf (tamp, namesize - 3, "%s%s", plan, videoName);

	snprintf (myPicture, 40, "%s%04d.%s", tamp, frame, ext);
	memcpy (Picture, myPicture, strlen (myPicture));
}
void
 getFileName (char *name, int format)
{
  char temp[80];
  char *myext[] = { "avi", "pnm", "jpg" };
  int i;
  time_t curdate;
  struct tm *tdate;
  memset (temp, '\0', sizeof (temp));
  time (&curdate);
  tdate = localtime (&curdate);
  snprintf (temp, 24, "%02d:%02d:%04d-%02d:%02d:%02d.%s\0",
	    tdate->tm_mon + 1, tdate->tm_mday, tdate->tm_year + 1900,
	    tdate->tm_hour, tdate->tm_min, tdate->tm_sec, myext[format]);

  memcpy (name, temp, strlen (temp));

}
static void
 getPicture( unsigned char *src,int width,int height, int frame)
{	FILE *foutpict;
	char myPicture[80];
	int k;
	unsigned char j;
	memset (myPicture,'\0',sizeof(myPicture));
	if (myPicture && (bpp == 3)) {
		getPictureName(myPicture,"-Fxxx","Picture",frame,"pnm");
	         foutpict = fopen (myPicture, "wb");
		fprintf (foutpict, "P6\n%d %d\n255\n", width, height);
		for (k = 0; k < width * height * bpp; k += 3) {
			j = src[k];
			src[k] = src[k + 2];
			src[k + 2] = j;
			}
		fwrite (src, sizeof (char), width * height * bpp, foutpict);
		fclose (foutpict);
	}
}

static void
 getPictureT( unsigned char *src,int width,int height, int frame)
{	FILE *foutpict;
	char myPicture[80];
	int k;
	unsigned char j;
	memset (myPicture,'\0',sizeof(myPicture));
	if (myPicture && (bpp == 3)) {
		getFileName(myPicture,1);
	         foutpict = fopen (myPicture, "wb");
		fprintf (foutpict, "P6\n%d %d\n255\n", width, height);
		for (k = 0; k < width * height * bpp; k += 3) {
			j = src[k];
			src[k] = src[k + 2];
			src[k + 2] = j;
			}
		fwrite (src, sizeof (char), width * height * bpp, foutpict);
		fclose (foutpict);
	}
}
#define CLIP(color) (unsigned char)((color>0xFF)?0xff:((color<0)?0:color))

static void
YUV420toRGB (unsigned char *src, unsigned char *dst, int width, int height,
	     int flipUV, int ColSpace)
{
	unsigned char *Y;
	unsigned char *V;
	unsigned char *U;
	int y1, y2, u, v;
	int v1, v2, u1, u2;
	unsigned char *pty1, *pty2;
	int i, j;
	unsigned char *RGB1, *RGB2;
	int r, g, b;

	//Initialization
	Y = src;
	V = Y + width * height;
	U = Y + width * height + width * height / 4;

	pty1 = Y;
	pty2 = pty1 + width;
	RGB1 = dst;
	RGB2 = RGB1 + 3 * width;
	for (j = 0; j < height; j += 2) {
		//printf ("process line %d\n",j);
		for (i = 0; i < width; i += 2) {
			if (flipUV) {
				u = (*V++) - 128;
				v = (*U++) - 128;
			} else {
				v = (*V++) - 128;
				u = (*U++) - 128;
			}
			switch (ColSpace) {
					// M$ color space
				case 0:{
						v1 = ((v << 10) + (v << 9) + (v << 6) + (v << 5)) >> 10;	// 1.593
						u1 = ((u << 8) + (u << 7) + (u << 4)) >> 10;	//         0.390
						v2 = ((v << 9) + (v << 4)) >> 10;	//                0.515
						u2 = ((u << 11) + (u << 4)) >> 10;	//               2.015
					}
					break;
					// PAL specific
				case 1:{
						v1 = ((v << 10) + (v << 7) + (v << 4)) >> 10;	//      1.1406
						u1 = ((u << 8) + (u << 7) + (u << 4) + (u << 3)) >> 10;	// 0.3984 
						v2 = ((v << 9) + (v << 6) + (v << 4) + (v << 1)) >> 10;	// 0.5800
						u2 = ((u << 11) + (u << 5)) >> 10;	//              2.0312
					}
					break;
					// V4l2
				case 2:{
						v1 = ((v << 10) + (v << 8) + (v << 7) + (v << 5)) >> 10;	//       1.406
						u1 = ((u << 8) + (u << 6) + (u << 5)) >> 10;	//                0.343
						v2 = ((v << 9) + (v << 7) + (v << 6) + (v << 5)) >> 10;	//        0.718
						u2 = ((u << 10) + (u << 9) + (u << 8) + (u << 4) + (u << 3)) >> 10;	// 1.773
					}
					break;
				case 3:{
						v1 = u1 = v2 = u2 = 0;
					}
					break;
				default:
					break;

			}
			//up-left

			y1 = (*pty1++);
			if (y1 > 0) {
				r = y1 + (v1);
				g = y1 - (u1) - (v2);
				b = y1 + (u2);

				r = CLIP (r);
				g = CLIP (g);
				b = CLIP (b);


			} else {
				r = g = b = 0;
			}
			*RGB1++ = r;
			*RGB1++ = g;
			*RGB1++ = b;

			//down-left


			y2 = (*pty2++);
			if (y2 > 0) {
				r = y2 + (v1);
				g = y2 - (u1) - (v2);
				b = y2 + (u2);

				r = CLIP (r);
				g = CLIP (g);
				b = CLIP (b);


			} else {
				r = b = g = 0;
			}
			*RGB2++ = r;
			*RGB2++ = g;
			*RGB2++ = b;

			//up-right
			y1 = (*pty1++);

			if (y1 > 0) {
				r = y1 + (v1);
				g = y1 - (u1) - (v2);
				b = y1 + (u2);

				r = CLIP (r);
				g = CLIP (g);
				b = CLIP (b);


			} else {
				r = g = b = 0;
			}

			*RGB1++ = r;
			*RGB1++ = g;
			*RGB1++ = b;



			//down-right
			y2 = (*pty2++);
			if (y2 > 0) {

				r = y2 + (v1);
				g = y2 - (u1) - (v2);
				b = y2 + (u2);

				r = CLIP (r);
				g = CLIP (g);
				b = CLIP (b);



			} else {
				r = b = g = 0;
			}

			*RGB2++ = r;
			*RGB2++ = g;
			*RGB2++ = b;

		}
		RGB1 += 3 * width;
		RGB2 += 3 * width;
		pty1 += width;
		pty2 += width;
	}

//printf ("done YUV420 -> RGB \n");
}
static void
refresh_screen (unsigned char *src, unsigned char *dst, int format, int width,
		int height, int size, int method)
{
	switch (format) {
		case VIDEO_PALETTE_JPEG:{
			libjpeg_decode (dst, src, size, width, height);
			}
			break;
		case VIDEO_PALETTE_RAW_JPEG:{
				if (method) {
				
				nolibjpeg_decode (dst, src, width, height);
				// compressed data decoder in userland
				
				} else {
				libjpeg_decode (dst, src, size, width, height);
				}
			}
			break;

		case VIDEO_PALETTE_YUV420P:{
				// uncompressed data yuv420P decoder in module
				YUV420toRGB ((unsigned char *) src,
					     (unsigned char *) dst, width,
					     height, 0, 0);

			}
			break;
		case VIDEO_PALETTE_RGB565:
		case VIDEO_PALETTE_RGB32:
		case VIDEO_PALETTE_RGB24:{
				// uncompressed data. rgb decoder in module simple copy for display.
				memcpy (dst, src, size);
			}
			break;
		default:
			break;
	}
}
static int
 isSpcaChip (const char *BridgeName)
{ 
	int i = -1;
	int size =0;
	size = strlen(BridgeName)-1;
	/* Spca506 return more with channel video, cut it */
	if (size > 10) size = 7;
	/* return Bridge otherwhise -1 */
	for (i=0; i < MAX_BRIDGE ;i++){
		if(strncmp(BridgeName,Blist[i].name,size) == 0) {
		 break;
		}
	}
	
return i;
}
static int
 getStreamId (const char * BridgeName)
{ 
	int i = -1;
	int match = -1;
/* return Stream_id otherwhise -1 */
	if((match=isSpcaChip(BridgeName)) < 0){
	 printf("Not an Spca5xx Camera !!\n");
	 return match;
	 }
	switch (match) {
		case BRIDGE_SPCA505:
		case BRIDGE_SPCA506: 
			i= YYUV;
			break;
		case BRIDGE_SPCA501:
			i = YUYV;
			break;
		case BRIDGE_SPCA508:
		 	i = YUVY;
			break;
		case BRIDGE_SPCA504:
		case BRIDGE_SPCA500:
		case BRIDGE_SPCA504B:
		case BRIDGE_SPCA533:
		case BRIDGE_SPCA504_PCCAM600:
		case BRIDGE_SPCA536:
		case BRIDGE_ZR364XX:
		case BRIDGE_ZC3XX:
		case BRIDGE_CX11646:
			i = JPEG;
			break;
		case BRIDGE_SONIX:
		case BRIDGE_SPCA561:
			i = GBRG;
			break;
		default:
			i = -1;
			 printf("Unable to find a StreamId !!\n");
			break;
	
	}
return i;	 
}
 
static int
 probeSize (const char *BridgeName, int *width, int *height)
{ 	int bridge = -1;
	int i;
	unsigned int intwidth;
	unsigned int intheight;
	unsigned int intformat;
	int match =0;
/* return 1->ok otherwhise -1 */
	if ((bridge= isSpcaChip(BridgeName)) < 0) {
	printf ("Cannot Probe Size !! maybe not an Spca5xx Camera\n");
	return -1;
	}
	
	for (i=0; (unsigned int)(GET_EXT_MODES(bridge)[i][0]);i++){
		intwidth = GET_EXT_MODES(bridge)[i][0];
		intheight = GET_EXT_MODES(bridge)[i][1];
		intformat = (GET_EXT_MODES(bridge)[i][2] & 0xF0) >> 4;
		if ((intwidth== *width) && (intheight == *height)){
			match = 1;
		} else {
			match = 0;
		}
		
		printf("Available Resolutions width %d  heigth %d %s %c\n",
			intwidth,intheight,(intformat)?"decoded":"native",
			(match)?'*':' ');
	}

}

/*
 * get_pic_mean:  Calculate the mean value of the pixels in an image.
 *                      This routine is used for adjusting the values of
 *                      the pixels to reasonable brightness.
 *
 * Arguments:           width, height = Dimensions of the picture
 *                      buffer    = Buffer to picture.
 *                      is_rgb    = 1 if the picture is rgb, else 0
 *                      start{x,y}, end{x,y} = Region to calculate the
 *                                  mean of.  This MUST be valid before
 *                                  being passed in!
 *
 * Return values:       Returns the average of all the components if rgb, else
 *                      the average whiteness of each pixel if B&W
 */
static int
 get_pic_mean( int width, int height, const unsigned char *buffer,
			  int is_rgb,int startx, int starty, int endx, 
			  int endy )
{
  double rtotal = 0, gtotal = 0, btotal = 0;
  int minrow, mincol, maxrow, maxcol, r, c;
  double bwtotal = 0, area;
  int rmean, gmean, bmean;
  const unsigned char *cp;
    
  minrow = starty;
  mincol = startx;
  maxrow = endy;
  maxcol = endx;

  area = (maxcol-mincol) * (maxrow-minrow);

  c = mincol;
  if( is_rgb ){
    for( r=minrow; r < maxrow; r++ ){
      cp = buffer + (r*width+c)*3;
      for( c=mincol; c < maxcol; c++ ) {
	rtotal += *cp++;
	gtotal += *cp++;
	btotal += *cp++;
      }
    }
    rmean = rtotal / area;
    gmean = gtotal / area;
    bmean = btotal / area;
    return (double)rmean * .299 +
	   (double)gmean * .587 +
           (double)bmean * .114;
  } else {
    for( r=minrow; r < maxrow; r++ ){
      cp = buffer + (r*width+c)*1;
      for( c=mincol; c < maxcol; c++ ) {
	bwtotal += *cp++;
      }
    }
    return (int)(bwtotal / area);
  }
}

static int
 adjust_bright( int width, int height, const char *addr, int is_rgb, struct video_picture *videopict, int fd)
{
  int totmean, newbright;
  
  totmean = get_pic_mean( width, height, addr, is_rgb, 0, 0,
				  width, height );
  newbright=videopict->brightness;
  if( totmean < BRIGHTMEAN - BRIGHTWINDOW|| 
      totmean > BRIGHTMEAN + BRIGHTWINDOW ){
    newbright += (BRIGHTMEAN-totmean)*SPRING_CONSTANT;
    newbright = clip_to(newbright, V4L_BRIGHT_MIN, V4L_BRIGHT_MAX);
  }
  
  printf("totmean:%03d newbright:%05d\r",totmean,newbright);
  /* Slight adjustment */
  videopict->brightness=newbright;
  setVideoPict(videopict,fd);
}

static int clip_to(int x, int low, int high){
  if(x<low)
      return low;
  else if (x>high)
      return high;
  else
      return x;
}
